#
#  TestRecord.py
#  JunkMatcher
#
#  Created by Benjamin Han on 2/1/05.
#  Copyright (c) 2005 Benjamin Han. All rights reserved.
#

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#!/usr/bin/env python

import threading  # to ensure thread-safety


class TestRecord (object):
    """Recording the past performance of a test
       ----------------------------------------
       time: the average CPU time spent on this property (in usec)
       
       numTruePositive: number of true positives
       numTrueNegative: number of true negatives
       numFalsePositive: number of false positives
       numFalseNegative: number of false negatives

       numTimes: total number of times this test has been invoked
       """
    # improving performance by not having __dict__
    __slots__ = ('_time', '_numTruePositive', '_numTrueNegative',
                 '_numFalsePositive', '_numFalseNegative', '_numTimes', 'lock')
    
    def __init__ (self, cpuTime = 0.0,
                  numTruePositive = 0, numTrueNegative = 0,
                  numFalsePositive = 0, numFalseNegative = 0):

        self._time = cpuTime
        self._numTruePositive = numTruePositive
        self._numTrueNegative = numTrueNegative
        self._numFalsePositive = numFalsePositive
        self._numFalseNegative = numFalseNegative
        self._numTimes = numTruePositive + numTrueNegative + numFalsePositive + numFalseNegative

        # it's not fine-grained enough; but most of the time we don't have overlapping read and
        # write access, so using a read-write lock might be an overkill
        self.lock = threading.Lock()

    def __str__ (self):
        self.lock.acquire()
        ret = '%f %d %d %d %d' % (self._time,
                                  self._numTruePositive, self._numTrueNegative,
                                  self._numFalsePositive, self._numFalseNegative)
        self.lock.release()
        return ret

    def reset (self):
        self.lock.acquire()
        self._time = 0.0
        self._numTruePositive = self._numTrueNegative = self._numFalsePositive = self._numFalseNegative = 0
        self._numTimes = 0
        self.lock.release()

    def setAll (self, cpuTime, numTruePositive, numTrueNegative,
                numFalsePositive, numFalseNegative):
        self.lock.acquire()
        
        self._time = cpuTime
        self._numTruePositive = numTruePositive
        self._numTrueNegative = numTrueNegative
        self._numFalsePositive = numFalsePositive
        self._numFalseNegative = numFalseNegative
        self._numTimes = numTruePositive + numTrueNegative + numFalsePositive + numFalseNegative

        self.lock.release()

    def addOne (self, isTrue, cpuTime, positive = True):
        """Adds one true/false result, and also updates the CPU time spent by this test.
        This is the default method called after a Property or a Pattern test is done 
        (because we assume the test result is correct unless told otherwise by a user).

        IMPORTANT: this should be called only ONCE - right after the corresponding test is
        done. Subsequent changes should be done via changeToFalse() and changeToTrue()."""
        self.lock.acquire()

        if isTrue:
            if positive: self._numTruePositive += 1
            else: self._numTrueNegative += 1            
        else:
            if positive: self._numFalsePositive += 1
            else: self._numFalseNegative += 1

        numTimes = self._numTimes + 1

        # re-averaging self._time
        if self._time == 0.0 or self._numTimes == 0:
            self._time = cpuTime
        else:
            self._time = (self._time * self._numTimes + cpuTime) / numTimes

        self._numTimes = numTimes
        self.lock.release()
        
    def changeToFalse (self, positive = True):
        """Since the default outcome for a test is always true, the only time
        this gets called is when the test result is later found false, hence we
        need to update numTruePositive and numTrueNegative as well."""
        self.lock.acquire()

        # the record has been reset: do nothing
        if self._numTimes == 0: return
        
        if positive:
            #assert self._numTruePositive >= 1
            self._numFalsePositive += 1
            self._numTruePositive -= 1
            
        else:
            #assert self._numTrueNegatove >= 1
            self._numFalseNegative += 1
            self._numTrueNegative -= 1

        self.lock.release()

    def changeToTrue (self, positive = True):
        """Similar in nature to changeToFalse(), but it's for the opposite case."""
        self.lock.acquire()

        # the record has been reset: do nothing
        if self._numTimes == 0: return
        
        if positive:
            #assert self._numFalsePositive >= 1
            self._numFalsePositive -= 1
            self._numTruePositive += 1
            
        else:
            #assert self._numFalseNegatove >= 1
            self._numFalseNegative -= 1
            self._numTrueNegative += 1

        self.lock.release()        

    def readTime (self):
        self.lock.acquire()
        ret = self._time
        self.lock.release()
        return ret

    def readNumbers (self):
        self.lock.acquire()
        ret = (self._numTruePositive, self._numTrueNegative,
               self._numFalsePositive, self._numFalseNegative)
        self.lock.release()
        return ret

    def readAll (self):
        self.lock.acquire()
        ret = (self._time,
               self._numTruePositive, self._numTrueNegative,
               self._numFalsePositive, self._numFalseNegative)
        self.lock.release()
        return ret

    def readTotal (self):
        self.lock.acquire()
        ret = self._numTimes
        self.lock.release()
        return ret

    def precision (self):
        self.lock.acquire()
        d = float(self._numTruePositive + self._numFalsePositive)
        if d != 0.0:
            p = self._numTruePositive / d
        else:
            p = 0.0            
        self.lock.release()
        return p

    def recall (self):
        self.lock.acquire()
        d = float(self._numTruePositive + self._numFalseNegative)
        if d != 0.0:
            r = self._numTruePositive / d
        else:
            r = 0.0            
        self.lock.release()
        return r
